#!/usr/bin/env python

##############################################################################
##
# This file is part of NavARP
##
# NavARP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# NavARP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
##
# You should have received a copy of the GNU General Public License
# along with NavARP.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python NavARP library. It defines the isomap
class which are extracted from a NavEntry."""

__author__ = ["Federico Bisti"]
__license__ = "GPL"
__date__ = "04/06/2021"

import numpy as np
import matplotlib.pyplot as plt
from scipy.ndimage import (
    gaussian_filter, gaussian_laplace, gaussian_gradient_magnitude)
import h5py

try:
    from navarp.utils import navplt, ktransf, isocut, kinterp
except ImportError:
    print("ImportError: navarp must be installed before using it.")


class IsoScan:
    """Sum of entry.data over the scan axis within [scan-dscan, scan+dscan].

    Args:
        entry (NavEntry class): The class for the data explored by NavARP;
        scan (float): scan value;
        dscan (float): delta scan value;
        norm_mode (string, optional, default='no'): if 'no', no
            normalization; if 'each', every Slice is normalized by the
            integral intensity of each Slice over their 90% of axis range;
            if "all", the overall z values are normalize to its min and
            max;
        sigma (array, optional, default=None): Standard deviation for
            Gaussian kernel. The standard deviations of the Gaussian filter
            are given for each axis as a sequence, or as a single number,
            in which case it is equal for all axes;
        order (integer, optional, default=2): The order of the filter along
            each axis is given as a sequence of integers, or as a single
            number. An order of 0 corresponds to convolution with a
            Gaussian kernel. A positive order corresponds to convolution
            with that derivative of a Gaussian. For example, an order of 2
            corresponds to convolution with that second derivative of a
            Gaussian;
        curvature (integer, optional, default=None): if not None the data
            are the curvature of the signal;
        kbins (integer, optional, default=None): if not None the data are
            interpolated on a uniform k-array binned with the kbins number.

    Attributes:
        angles (ndarray): Angular axis of the analyzer;
        data (ndarray): The matrix composed by the detector images;
        data_init (ndarray): The matrix composed by the detector images without
            any normalization;
        dscan (float): delta scan value;
        ebins (ndarray): Binding energy axis;
        efermi (ndarray): Fermi level kinetic energy (or energies);
        ekins (ndarray): Kinetic energy axis of the analyzer;
        hv (ndarray): Photon energy without any reshape, it is present only in
            the case of hv scan;
        scans (float): scan value;
        kx (ndarray): k vector along the slit.
    """

    def __init__(
        self,
        entry,
        scan,
        dscan=0,
        norm_mode='no',
        sigma=None,
        order=2,
        curvature=None,
        kbins=None
    ):
        """Class initialization."""
        self.scan = scan
        self.dscan = dscan
        self.data_init = isocut.maps_sum(
            scan, dscan, entry.scans, entry.data)

        self.ebins = entry.ebins
        if (entry.scan_type == "hv") or (entry.scan_type == "repeated"):
            scan_ind = np.argmin(abs(entry.scans-scan))
            self.ekins = entry.energies[scan_ind, :]
            self.efermi = entry.efermi[scan_ind]
            self.hv = entry.hv[scan_ind]
        else:
            self.ekins = entry.energies
            self.efermi = entry.efermi
            self.hv = entry.hv

        self.angles = entry.angles
        self.kx_interp = None
        if entry.tht_an is not None:
            self.kx = ktransf.get_k_along_slit(
                e_kins=self.ekins,
                tht=self.angles,
                tht_an=entry.tht_an,
                p_hv=entry.p_hv,
                hv=self.hv,
                tht_ap=entry.analyzer.tht_ap
            )

        self.dataprocessing(norm_mode, sigma, order, curvature, kbins)

    def get_axs_and_data(self, xname='auto', yname='auto'):
        """Return the data and relative axes.

        Args:
            xname (string, optional, auto/tht/kx): Select the x variable;
            yname (string, optional, auto/eef/ekin): Select the y variable.
        Returns:
            xax, yax, data, xname, yname
        """
        if self.kx_interp is not None:
            xname = 'kx'
            xax = self.kx_interp

        else:
            if xname == 'auto':
                if hasattr(self, 'kx'):
                    xname = 'kx'
                    xax = self.kx
                else:
                    xname = 'tht'
                    xax = self.angles
            elif xname == 'tht':
                xax = self.angles
            elif xname == 'kx':
                xax = self.kx

        if yname == 'auto':
            yname = 'eef'
            yax = self.ebins
        elif yname == 'ekin':
            yax = self.ekins
        elif yname == 'eef':
            yax = self.ebins

        return xax, yax, self.data, xname, yname

    def dataprocessing(
        self,
        norm_mode='no',
        sigma=None,
        order=2,
        curvature=None,
        kbins=None
    ):
        """Apply data post-processing procedure.

        Args:
            norm_mode (string, optional, default='no'): if 'no', no
                normalization; if 'each', every Slice is normalized by the
                integral intensity of each Slice over their 90% of axis range;
                if 'all', the overall z values are normalize to its min and
                max;
            sigma (array, optional, default=None): Standard deviation for
                Gaussian kernel. The standard deviations of the Gaussian filter
                are given for each axis as a sequence, or as a single number,
                in which case it is equal for all axes;
            order (integer, optional, default=2): The order of the filter along
                each axis is given as a sequence of integers, or as a single
                number. An order of 0 corresponds to convolution with a
                Gaussian kernel. A positive order corresponds to convolution
                with that derivative of a Gaussian. For example, an order of 2
                corresponds to convolution with that second derivative of a
                Gaussian;
            curvature (float, optional, default=None): if not None the data
                are the curvature of the signal, only in 1-dimension, meaning
                only for sigma with one value;
            kbins (integer, optional, default=None): if not None the data are
                interpolated on a uniform k-array binned with the kbins number.
        """
        if kbins is not None:
            if hasattr(self, 'kx'):
                self.kx_interp, self.data = kinterp.get_isoscan(
                    self.kx, self.ebins, self.data_init, kbins)
            else:
                raise AttributeError('No valid tht_an, do set_tht_an first')
        else:
            self.data = np.copy(self.data_init)
            self.kx_interp = None

        if norm_mode == 'each':
            self.data = navplt.norm(self.data, mode=norm_mode, axis=1)

        if sigma is not None:
            self.data = postprocessing(self.data, sigma, order, curvature)

        if norm_mode == 'all':
            self.data = navplt.norm(self.data, mode=norm_mode, axis=1)

    def show(
        self,
        xname='auto',
        yname='auto',
        cmap='magma_r',
        ax=None,
        z_range=None,
        style=None,
        print_out=False,
        cmapscale="linear"
    ):
        """Plot the isomap.

        Args:
            xname (string, optional, auto/tht/kx): Select the x variable;
            yname (string, optional, auto/eef/ekin): Select the y variable;
            cmap (string, optional): Colormap;
            ax (matplotlib.axes, optional): Axes, if None it is created inside
                the function;
            z_range (array, optional, [min, max]): Extreme values for z scale;
            style (string, optional): See navplt.give_style function;
            print_out (boolean, optional): If True, print the axes dimension
                with them adaptations;
            cmapscale (string, optional, linear/log/power): Select the scale
                mode.
        Returns:
            The returned value is the object matplotlib.collections.QuadMesh
                from ax.pcolormesh.
        """
        xax, yax, data, xname, yname = self.get_axs_and_data(xname, yname)

        if style is None:
            style = xname + '_' + yname

        qmisoscan = navplt.pimage(
            xax,
            yax,
            data,
            cmap=cmap,
            ax=ax,
            z_range=z_range,
            style=style,
            print_out=print_out,
            cmapscale=cmapscale
        )
        return qmisoscan

    def export_as_hdf5(
        self,
        file_path,
        xname='auto',
        yname='auto',
        title='NavARP_isoscan'
    ):
        """Save the isomap as hdf5 format.

        Args:
            xname (string, optional, auto/tht/kx): Select the x variable;
            yname (string, optional, auto/eef/ekin): Select the y variable;
            title (string, optiona): data title.
        """
        xax, yax, data, xname, yname = self.get_axs_and_data(xname, yname)

        f = h5py.File(file_path, "w")  # create the HDF5 file

        nxdata = f.create_group(u"isoscan")

        nxdata.create_dataset(u"title", data=title)

        if xname == 'kx':
            axes_names = ['momentum', '']
            xdata = nxdata.create_dataset(axes_names[0], data=xax)
            xdata.attrs[u"units"] = u"1/angstrom"
        else:
            axes_names = ['angles', '']
            xdata = nxdata.create_dataset(axes_names[0], data=xax)
            xdata.attrs[u"units"] = u"degrees"
        xdata.attrs[u"axis"] = 1
        xdata.attrs[u"primary"] = 1

        if yname == 'eef':
            axes_names[1] = "binding_energies"
        else:
            axes_names[1] = "kinetic_energies"
        ydata = nxdata.create_dataset(axes_names[1], data=yax)
        ydata.attrs[u"axis"] = 2
        ydata.attrs[u"primary"] = 1
        ydata.attrs[u"units"] = u"eV"

        data_nx = nxdata.create_dataset(
            u"data", data=data, compression='gzip', chunks=True)
        data_nx.attrs[u"units"] = u"counts"
        nxdata.attrs[u"signal"] = u"data"

        nxdata.attrs[u"axes"] = [axesn.encode('utf8') for axesn in axes_names]

        f.close()

    def export_as_nxs(
        self,
        file_path,
        xname='auto',
        yname='auto',
        title='NavARP_isoscan'
    ):
        """Save the isomap as NXdata nexus class.

        Args:
            xname (string, optional, auto/tht/kx): Select the x variable;
            yname (string, optional, auto/eef/ekin): Select the y variable;
            title (string, optiona): data title.
        """
        xax, yax, data, xname, yname = self.get_axs_and_data(xname, yname)

        if len(xax.shape) > 1:
            raise ValueError(
                "Data must be interpolated on uniform grid, first. "
                "Use kbins during isoscan initialization or in dataprocessing."
            )
            return None

        f = h5py.File(file_path, "w")  # create the HDF5 NeXus file
        f.attrs[u"default"] = u"isoscan"

        nxdata = f.create_group(u"isoscan")
        nxdata.attrs[u"NX_class"] = u"NXdata"

        nxdata.create_dataset(u"title", data=title)

        if xname == 'kx':
            axes_names = ['', 'momentum']
            xdata = nxdata.create_dataset(axes_names[1], data=xax)
            xdata.attrs[u"units"] = u"1/angstrom"
        else:
            axes_names = ['', 'angles']
            xdata = nxdata.create_dataset(axes_names[1], data=xax)
            xdata.attrs[u"units"] = u"degrees"
        xdata.attrs[u"axis"] = 2
        xdata.attrs[u"primary"] = 1

        if yname == 'eef':
            axes_names[0] = "binding_energies"
        else:
            axes_names[0] = "kinetic_energies"
        ydata = nxdata.create_dataset(axes_names[0], data=yax)
        ydata.attrs[u"axis"] = 1
        ydata.attrs[u"primary"] = 1
        ydata.attrs[u"units"] = u"eV"

        data_nx = nxdata.create_dataset(
            u"data", data=data.transpose(), compression='gzip', chunks=True)
        data_nx.attrs[u"units"] = u"counts"
        nxdata.attrs[u"signal"] = u"data"

        nxdata.attrs[u"axes"] = [axesn.encode('utf8') for axesn in axes_names]

        f.close()

    def export_as_itx(
        self,
        file_path,
        xname='auto',
        yname='auto',
        wave_name='NavARP_isoscan'
    ):
        """Save the isomap as Igor Pro Text file.

        Args:
            xname (string, optional, auto/tht/kx): Select the x variable;
            yname (string, optional, auto/eef/ekin): Select the y variable;
            wave_name (string, optiona): name of the wave.
        """
        xax, yax, data, xname, yname = self.get_axs_and_data(xname, yname)

        if len(xax.shape) > 1:
            raise ValueError(
                "Data must be interpolated on uniform grid, first. "
                "Use kbins during isoscan initialization or in dataprocessing."
            )
            return None

        if xname == 'kx':
            xlabel = 'Momentum [1/Angstrom]'
        else:
            xlabel = 'Angles [degree]'

        if yname == 'eef':
            ylabel = "Binding Energy [eV]"
        else:
            ylabel = "Kinetic Energy [eV]"

        header = (
            "IGOR\n" +
            "WAVES/N=({0[0]},{0[1]})\t".format(data.shape) +
            wave_name + "\n" + "BEGIN"
        )

        footer = (
            'END\n' +
            'X SetScale/I x {},{},"{}", {};\n'.format(
                xax[0], xax[-1], xlabel, wave_name) +
            'X SetScale/I y {},{},"{}", {};\n'.format(
                yax[0], yax[-1], ylabel, wave_name) +
            'X SetScale d 0,0,"Counts [arb. units]", {};\n'.format(wave_name) +
            'X Note {}, "Isoscan at {} +- {}, generated by NavARP"'.format(
                wave_name, self.scan, self.dscan)
        )

        np.savetxt(
            file_path,
            data,
            fmt='%.6f',
            delimiter=' ',
            newline='\n',
            header=header,
            footer=footer,
            comments='',
            encoding=None
        )


class IsoEnergy:
    """Sum of entry.data over the ebin axis within [ebin-debin, ebin+debin].

    Args:
        entry (NavEntry class): The class for the data explored by NavARP;
        ebin (float): Binding energy value;
        debin (float): Delta binding energy value;
        norm_mode (string, optional, default='no'): if 'no', no
            normalization; if 'each', every Slice is normalized by the
            integral intensity of each Slice over their 90% of axis range;
            if "all", the overall z values are normalize to its min and
            max;
        sigma (array, optional, default=None): Standard deviation for
            Gaussian kernel. The standard deviations of the Gaussian filter
            are given for each axis as a sequence, or as a single number,
            in which case it is equal for all axes;
        order (integer, optional, default=2): The order of the filter along
            each axis is given as a sequence of integers, or as a single
            number. An order of 0 corresponds to convolution with a
            Gaussian kernel. A positive order corresponds to convolution
            with that derivative of a Gaussian. For example, an order of 2
            corresponds to convolution with that second derivative of a
            Gaussian;
        curvature (integer, optional, default=None): if not None the data
            are the curvature of the signal;
        kbins (array, optional, default=None): if not None the data are
            interpolated on a uniform k-array binned with the kbins number.

    Attributes:
        angles (ndarray): Angular axis of the analyzer;
        data (ndarray): The matrix composed by the detector images;
        data_init (ndarray): The matrix composed by the detector images without
            any normalization;
        debin (float): Delta binding energy value;
        ebin (ndarray): Binding energy value;
        hv (ndarray): Photon energy without any reshape, it is present only in
            the case of hv scan;
        ks (ndarray): k vector along the direction consistent with the scan
            axis;
        kx (ndarray): k vector along the slit;
        scan_type (str): Acquisition method;
        scans (ndarray): Scan axis of the acquisition method.
    """

    def __init__(
        self,
        entry,
        ebin,
        debin=0,
        norm_mode='no',
        sigma=None,
        order=2,
        curvature=None,
        kbins=None
    ):
        """Class initialization."""
        self.ebin = ebin
        self.debin = debin
        self.data_init = isocut.maps_sum(
            ebin, debin, entry.ebins, entry.data, axis=2)

        self.scan_type = entry.scan_type

        if self.scan_type == 'hv':
            self.scans = entry.hv
        else:
            self.scans = entry.scans

        self.hv = entry.hv
        self.angles = entry.angles
        if entry.valid_kspace:
            e_kin_val = ebin + entry.efermi
            self.kx, self.ks = ktransf.get_k_isoen(entry, e_kin_val)

        self.dataprocessing(norm_mode, sigma, order, curvature, kbins)

    def dataprocessing(
        self,
        norm_mode='no',
        sigma=None,
        order=2,
        curvature=None,
        kbins=None
    ):
        """Apply data post-processing procedure.

        Args:
            norm_mode (string, optional, default='no'): if 'no', no
                normalization; if 'each', every Slice is normalized by the
                integral intensity of each Slice over their 90% of axis range;
                if 'all', the overall z values are normalize to its min and
                max;
            sigma (array, optional, default=None): Standard deviation for
                Gaussian kernel. The standard deviations of the Gaussian filter
                are given for each axis as a sequence, or as a single number,
                in which case it is equal for all axes;
            order (integer, optional, default=2): The order of the filter along
                each axis is given as a sequence of integers, or as a single
                number. An order of 0 corresponds to convolution with a
                Gaussian kernel. A positive order corresponds to convolution
                with that derivative of a Gaussian. For example, an order of 2
                corresponds to convolution with that second derivative of a
                Gaussian;
            curvature (float, optional, default=None): if not None the data
                are the curvature of the signal, only in 1-dimension, meaning
                only for sigma with one value;
            kbins (array, optional, default=None): if not None, the data are
                interpolated on a uniform k-array binned following the
                [kxbins, kybins] numbers; kbins must be an array of two integer
                [kxbins, kybins] not equal; if from the input kxbins == kybins
                then kybins = kybins + 1 to make it different from kxbins (this
                is necessary due to how the code recognize the element order
                and placement in the data matrix).
        """
        if kbins is not None:
            if len(kbins) != 2:
                raise ValueError(
                    'No valid kbins, it must be an array of two integers')

            if hasattr(self, 'kx'):
                if kbins[0] == kbins[1]:
                    kbins[1] += 1
                self.kx_interp, self.ks_interp, self.data = kinterp.get_isoen(
                    kbins[0], kbins[1], self.kx, self.ks, self.data_init)
                if self.ks_interp.shape[0] == self.data.shape[1]:
                    self.data = self.data.transpose()
            else:
                raise AttributeError('No valid tht_an, do set_tht_an first')
        else:
            self.data = np.copy(self.data_init)
            self.kx_interp = None
            self.ks_interp = None

        if norm_mode == 'each':
            self.data = navplt.norm(self.data, mode=norm_mode, axis=1)

        if sigma is not None:
            self.data = postprocessing(self.data, sigma, order, curvature)

        if norm_mode == 'all':
            self.data = navplt.norm(self.data, mode=norm_mode, axis=1)

    def get_axs_and_data(self, xname='auto', yname='auto'):
        """Return the data and relative axes.

        Args:
            xname (string, optional, auto/tht/kx): Select the x variable;
            yname (string, optional, auto/eef/ekin): Select the y variable.
        Returns:
            xax, yax, data, xname, yname
        """
        if self.kx_interp is not None:
            xname = 'kx'
            xax = self.kx_interp

            if self.scan_type == 'hv':
                yname = 'kz'
            else:
                yname = 'ky'
            yax = self.ks_interp
        else:
            if xname == 'auto':
                if hasattr(self, 'kx'):
                    xname = 'kx'
                    xax = self.kx
                else:
                    xname = 'tht'
                    xax = self.angles
            elif xname == 'tht':
                xax = self.angles
            elif xname == 'kx':
                xax = self.kx
            else:
                xax = self.angles

            if yname == 'auto' or (yname == 'ks'):
                if hasattr(self, 'ks'):
                    if self.scan_type == 'hv':
                        yname = 'kz'
                    else:
                        yname = 'ky'
                    yax = self.ks
                else:
                    if self.scan_type == 'hv':
                        yname = 'hv'
                    else:
                        yname = 'phi'
                    yax = self.scans
            elif (yname == 'phi') or (yname == 'hv'):
                yax = self.scans
            elif (yname == 'ky') or (yname == 'kz'):
                yax = self.ks
            else:
                yax = self.scans

        return xax, yax, self.data, xname, yname

    def show(
        self,
        xname='auto',
        yname='auto',
        cmap='magma_r',
        ax=None,
        z_range=None,
        style=None,
        print_out=False,
        cmapscale="linear",
        rotate=None
    ):
        """Plot the isomap.

        Args:
            xname (string, optional, auto/tht/kx): Select the x variable;
            yname (string, optional, auto/phi/hv/ky/kz): Select the y variable;
            cmap (string, optional): Colormap;
            ax (matplotlib.axes, optional): Axes, if None it is created inside
                the function;
            z_range (array, optional, [min, max]): Extreme values for z scale;
            style (string, optional): See navplt.give_style function;
            print_out (boolean, optional): If True, print the axes dimension
                with them adaptations;
            cmapscale (string, optional, linear/log/power): Select the scale
                mode;
            rotate (float, optional): angle in degrees for a counter clockwise
                rotation of the shown isoenergy.
        Returns:
            The returned value is the object matplotlib.collections.QuadMesh
                from ax.pcolormesh.
        """
        xax, yax, data, xname, yname = self.get_axs_and_data(xname, yname)

        if style is None:
            style = xname + '_' + yname

        if rotate:
            if len(xax.shape) != 2:
                xax_cp = np.tile(xax, (self.scans.shape[0], 1))
            else:
                xax_cp = np.copy(xax)

            if len(yax.shape) != 2:
                yax_cp = np.tile(yax, (self.angles.shape[0], 1)).T
            else:
                yax_cp = np.copy(yax)

            rot_ang = np.deg2rad(rotate)
            xax = np.cos(rot_ang)*xax_cp - np.sin(rot_ang)*yax_cp
            yax = np.sin(rot_ang)*xax_cp + np.cos(rot_ang)*yax_cp

        qmiso = navplt.pimage(
            xax,
            yax,
            data,
            cmap=cmap,
            ax=ax,
            z_range=z_range,
            style=style,
            print_out=print_out,
            cmapscale=cmapscale
        )
        return qmiso

    def export_as_hdf5(
        self,
        file_path,
        xname='auto',
        yname='auto',
        title='NavARP_isoenergy'
    ):
        """Save the isomap as hdf5 format.

        Args:
            xname (string, optional, auto/tht/kx): Select the x variable;
            yname (string, optional, auto/eef/ekin): Select the y variable;
            title (string, optiona): data title.
        """
        xax, yax, data, xname, yname = self.get_axs_and_data(xname, yname)

        f = h5py.File(file_path, "w")  # create the HDF5 NeXus file

        nxdata = f.create_group(u"isoenergy")

        nxdata.create_dataset(u"title", data=title)

        if xname == 'kx':
            axes_names = ['', 'k_along_slit']
            xdata = nxdata.create_dataset(axes_names[1], data=xax)
            xdata.attrs[u"units"] = u"1/angstrom"
        else:
            axes_names = ['', 'angles']
            xdata = nxdata.create_dataset(axes_names[1], data=xax)
            xdata.attrs[u"units"] = u"degrees"
        xdata.attrs[u"axis"] = 2
        xdata.attrs[u"primary"] = 1

        if (yname == 'ky') or (yname == 'kz'):
            axes_names[0] = 'k_along_scan'
            ydata = nxdata.create_dataset(axes_names[0], data=yax)
            ydata.attrs[u"units"] = u"1/angstrom"
        else:
            axes_names[0] = self.scan_type
            ydata = nxdata.create_dataset(axes_names[0], data=yax)
            if self.scan_type == 'hv':
                ydata.attrs[u"units"] = u"eV"
            else:
                ydata.attrs[u"units"] = u"degree"
        ydata.attrs[u"axis"] = 1
        ydata.attrs[u"primary"] = 1

        data_nx = nxdata.create_dataset(
            u"data", data=data, compression='gzip', chunks=True)
        data_nx.attrs[u"units"] = u"counts"
        nxdata.attrs[u"signal"] = u"data"

        nxdata.attrs[u"axes"] = axes_names

        f.close()

    def export_as_nxs(
        self,
        file_path,
        xname='auto',
        yname='auto',
        title='NavARP_isoenergy'
    ):
        """Save the isomap as NXdata nexus class.

        Args:
            xname (string, optional, auto/tht/kx): Select the x variable;
            yname (string, optional, auto/eef/ekin): Select the y variable;
            title (string, optiona): data title.
        """
        xax, yax, data, xname, yname = self.get_axs_and_data(xname, yname)

        if (len(xax.shape) > 1) or (len(yax.shape) > 1):
            raise ValueError(
                "Data must be interpolated on uniform grid, first. Use "
                "kbins during isoenergy initialization or in dataprocessing."
            )
            return None

        f = h5py.File(file_path, "w")  # create the HDF5 NeXus file
        f.attrs[u"default"] = u"isoenergy"

        nxdata = f.create_group(u"isoenergy")
        nxdata.attrs[u"NX_class"] = u"NXdata"

        nxdata.create_dataset(u"title", data=title)

        if xname == 'kx':
            axes_names = [u'', u'k_along_slit']
            xdata = nxdata.create_dataset(axes_names[1], data=xax)
            xdata.attrs[u"units"] = u"1/angstrom"
        else:
            axes_names = [u'', u'angles']
            xdata = nxdata.create_dataset(axes_names[1], data=xax)
            xdata.attrs[u"units"] = u"degrees"
        xdata.attrs[u"axis"] = 2
        xdata.attrs[u"primary"] = 1

        if (yname == 'ky') or (yname == 'kz'):
            axes_names[0] = u'k_along_scan'
            ydata = nxdata.create_dataset(axes_names[0], data=yax)
            ydata.attrs[u"units"] = u"1/angstrom"
        else:
            axes_names[0] = self.scan_type
            ydata = nxdata.create_dataset(axes_names[0], data=yax)
            if self.scan_type == u'hv':
                ydata.attrs[u"units"] = u"eV"
            else:
                ydata.attrs[u"units"] = u"degree"
        ydata.attrs[u"axis"] = 1
        ydata.attrs[u"primary"] = 1

        data_nx = nxdata.create_dataset(
            u"data", data=data, compression='gzip', chunks=True)
        data_nx.attrs[u"units"] = u"counts"
        nxdata.attrs[u"signal"] = u"data"

        nxdata.attrs[u"axes"] = axes_names

        f.close()

    def export_as_itx(
        self,
        file_path,
        xname='auto',
        yname='auto',
        wave_name='NavARP_isoenergy'
    ):
        """Save the isomap as Igor Pro Text file.

        Args:
            xname (string, optional, auto/tht/kx): Select the x variable;
            yname (string, optional, auto/eef/ekin): Select the y variable;
            wave_name (string, optiona): name of the wave.
        """
        xax, yax, data, xname, yname = self.get_axs_and_data(xname, yname)

        if (len(xax.shape) > 1) or (len(yax.shape) > 1):
            raise ValueError(
                "Data must be interpolated on uniform grid, first. Use "
                "kbins during isoenergy initialization or in dataprocessing."
            )
            return None

        # data transpose for itx
        data_itx = data.transpose()

        if xname == 'kx':
            xlabel = 'k along slit [1/Angstrom]'
        else:
            xlabel = 'Angles [degree]'

        if (yname == 'ky') or (yname == 'kz'):
            ylabel = 'k along scan [1/Angstrom]'
        else:
            if self.scan_type == 'hv':
                ylabel = 'Photon energy [eV]'
            else:
                ylabel = '{} [degree]'.format(self.scan_type)

        header = (
            "IGOR\n" +
            "WAVES/N=({0[0]},{0[1]})\t".format(data_itx.shape) +
            wave_name + "\n" + "BEGIN"
        )

        footer = (
            'END\n' +
            'X SetScale/I x {},{},"{}", {};\n'.format(
                xax[0], xax[-1], xlabel, wave_name) +
            'X SetScale/I y {},{},"{}", {};\n'.format(
                yax[0], yax[-1], ylabel, wave_name) +
            'X SetScale d 0,0,"Counts [arb. units]", {};\n'.format(wave_name) +
            'X Note {}, "isoenergy at {} +- {}, generated by NavARP"'.format(
                wave_name, self.ebin, self.debin)
        )

        np.savetxt(
            file_path,
            data_itx,
            fmt='%.6f',
            delimiter=' ',
            newline='\n',
            header=header,
            footer=footer,
            comments='',
            encoding=None
        )


class IsoAngle:
    """Sum of entry.data over the angle axis in [angle-dangle, angle+dangle].

    Args:
        entry (NavEntry class): The class for the data explored by NavARP;
        angle (float): Angular value on the;
        dangle (float): Delta angular value;
        norm_mode (string, optional, default='no'): if 'no', no
            normalization; if 'each', every Slice is normalized by the
            integral intensity of each Slice over their 90% of axis range;
            if "all", the overall z values are normalize to its min and
            max;
        sigma (array, optional, default=None): Standard deviation for
            Gaussian kernel. The standard deviations of the Gaussian filter
            are given for each axis as a sequence, or as a single number,
            in which case it is equal for all axes;
        order (integer, optional, default=2): The order of the filter along
            each axis is given as a sequence of integers, or as a single
            number. An order of 0 corresponds to convolution with a
            Gaussian kernel. A positive order corresponds to convolution
            with that derivative of a Gaussian. For example, an order of 2
            corresponds to convolution with that second derivative of a
            Gaussian;
        curvature (integer, optional, default=None): if not None the data
            are the curvature of the signal.

    Attributes:
        angle (float): Angular value on the;
        dangle (float): Delta angular value;
        data (ndarray): The matrix composed by the detector images;
        data_init (ndarray): The matrix composed by the detector images without
            any normalization;
        ebins (ndarray): Binding energy axis;
        efermi (ndarray): Fermi level kinetic energy (or energies);
        ekins (ndarray): Kinetic energy axis of the analyzer;
        hv (ndarray): Photon energy without any reshape, it is present only in
            the case of hv scan;
        scan_type (str): Acquisition method;
        scans (ndarray): Scan axis of the acquisition method.
    """

    def __init__(
        self,
        entry,
        angle,
        dangle=0,
        norm_mode='no',
        sigma=None,
        order=2,
        curvature=None,
        kbins=None
    ):
        """Class initialization."""
        self.angle = angle
        self.dangle = dangle
        self.data_init = isocut.maps_sum(
            angle, dangle, entry.angles, entry.data)

        self.ekins = entry.energies
        self.ebins = entry.ebins
        self.efermi = entry.efermi
        self.hv = entry.hv

        self.scan_type = entry.scan_type

        if self.scan_type == 'hv':
            self.scans = entry.hv
        else:
            self.scans = entry.scans

        self.dataprocessing(norm_mode, sigma, order, curvature)

    def show(
        self,
        xname='auto',
        yname='auto',
        cmap='magma_r',
        ax=None,
        z_range=None,
        style=None,
        print_out=False,
        cmapscale="linear"
    ):
        """Plot the isomap.

        Args:
            xname (string, optional, auto/phi/hv): Select the x variable;
            yname (string, optional, auto/eef/ekin): Select the y variable;
            cmap (string, optional): Colormap;
            ax (matplotlib.axes, optional): Axes, if None it is created inside
                the function;
            z_range (array, optional, [min, max]): Extreme values for z scale;
            style (string, optional): See navplt.give_style function;
            print_out (boolean, optional): If True, print the axes dimension
                with them adaptations;
            cmapscale (string, optional, linear/log/power): Select the scale
                mode.
        Returns:
            The returned value is the object matplotlib.collections.QuadMesh
                from ax.pcolormesh.

        """
        if xname == 'auto' or (xname == 'scan'):
            if self.scan_type == 'hv':
                xname = 'hv'
            else:
                xname = 'phi'
            x_ax = self.scans
        elif xname == 'ekin':
            x_ax = self.ekins
        elif xname == 'eef':
            x_ax = self.ebins

        if yname == 'auto':
            yname = 'eef'
            y_ax = self.ebins
        elif yname == 'ekin':
            y_ax = self.ekins
        elif yname == 'eef':
            y_ax = self.ebins
        elif yname == 'scan':
            if self.scan_type == 'hv':
                yname = 'hv'
            else:
                yname = 'phi'
            y_ax = self.scans

        if style is None:
            style = xname + '_' + yname

        qmisoangle = navplt.pimage(
            x_ax,
            y_ax,
            self.data,
            cmap=cmap,
            ax=ax,
            z_range=z_range,
            style=style,
            print_out=print_out,
            cmapscale=cmapscale
        )
        return qmisoangle

    def dataprocessing(
        self,
        norm_mode='no',
        sigma=None,
        order=2,
        curvature=None
    ):
        """Use data post-processing procedure.

        Args:
            norm_mode (string, optional, default='no'): if 'no', no
                normalization; if 'each', every Slice is normalized by the
                integral intensity of each Slice over their 90% of axis range;
                if 'all', the overall z values are normalize to its min and
                max;
            sigma (array, optional, default=None): Standard deviation for
                Gaussian kernel. The standard deviations of the Gaussian filter
                are given for each axis as a sequence, or as a single number,
                in which case it is equal for all axes;
            order (integer, optional, default=2): The order of the filter along
                each axis is given as a sequence of integers, or as a single
                number. An order of 0 corresponds to convolution with a
                Gaussian kernel. A positive order corresponds to convolution
                with that derivative of a Gaussian. For example, an order of 2
                corresponds to convolution with that second derivative of a
                Gaussian;
            curvature (integer, optional, default=None): if not None the data
                are the curvature of the signal.
        """
        self.data = np.copy(self.data_init)

        if norm_mode == 'each':
            self.data = navplt.norm(self.data, mode=norm_mode, axis=1)

        if sigma is not None:
            self.data = postprocessing(self.data, sigma, order, curvature)

        if norm_mode == 'all':
            self.data = navplt.norm(self.data, mode=norm_mode, axis=1)


class IsoK:
    """Interpolate the entry.data along specified k-points.

    Args:
        entry (NavEntry class): the class for the data explored by NavARP;
        kx_pts (ndarray): kx coordinate of the points defining the path;
        ky_pts (ndarray): ky coordinate of the points defining the path;
        klabels (list of string, optional): list of the labels for each point
            in the kspace;
        ebins_interp (ndarray, optional): The binding energies vector to sample
            the entry.data at, if None then it is entry.ebins;
        kbins (int, optional): defines the number of equal-width bins in the
            given path, if None then it is calculated from kx;
        mask_once (boolean, optional): If True, the region (mask), where the
            interpolation function is going to be defined (the smaller the
            faster is the interpolation procedure), is generated once and used
            in all the ebins_interp range;
        norm_mode (string, optional, default='no'): if 'no', no
            normalization; if 'each', every Slice is normalized by the
            integral intensity of each Slice over their 90% of axis range;
            if "all", the overall z values are normalize to its min and
            max;
        sigma (array, optional, default=None): Standard deviation for
            Gaussian kernel. The standard deviations of the Gaussian filter
            are given for each axis as a sequence, or as a single number,
            in which case it is equal for all axes;
        order (integer, optional, default=2): The order of the filter along
            each axis is given as a sequence of integers, or as a single
            number. An order of 0 corresponds to convolution with a
            Gaussian kernel. A positive order corresponds to convolution
            with that derivative of a Gaussian. For example, an order of 2
            corresponds to convolution with that second derivative of a
            Gaussian;
        curvature (float, optional, default=None): if not None the data
            are the curvature of the signal, only in 1-dimension, meaning
            only for sigma with one value.

    Attributes:
        data_init (ndarray): The matrix composed by interpolated values along
            the path without any normalization;
        data (ndarray): The matrix composed by interpolated values along the
            path with the normalization;
        ebins (ndarray): Binding energy axis;
        efermi (ndarray): Fermi level kinetic energy (or energies);
        ekins (ndarray): Kinetic energy axis of the analyzer (defined only if
            scan is not on the photon energy);
        hv (ndarray): Photon energy without any reshape, it is present only in
            the case of hv scan;
        krho (ndarray): the distance vector from the first point;
        kxy_interp (ndarray): the (kx, ky) points along the path;
        kx_pts (ndarray): kx coordinate of the points defining the path;
        ky_pts (ndarray): ky coordinate of the points defining the path;
        klabels (list of string): Array of the labels for each k_ptys_xy;
        k_pts_bin (ndarray of int): the index values defining the segments
            along the path;
        scan_type (str): Acquisition method.
    """

    def __init__(
        self,
        entry,
        kx_pts,
        ky_pts,
        klabels=None,
        ebins_interp=None,
        kbins=None,
        mask_once=True,
        norm_mode='no',
        sigma=None,
        order=2,
        curvature=None
    ):
        """Class initialization."""
        self.klabels = klabels
        self.kx_pts = kx_pts
        self.ky_pts = ky_pts

        [self.krho,
         self.kxy_interp,
         self.ebins_interp,
         self.data_init,
         self.k_pts_bin] = kinterp.get_isok(
            entry, kx_pts, ky_pts, ebins_interp, kbins, mask_once)

        self.ebins = self.ebins_interp
        self.efermi = entry.efermi

        self.hv = entry.hv

        self.scan_type = entry.scan_type

        if self.scan_type == 'hv':
            self.scans = entry.hv
        else:
            self.scans = entry.scans
            self.ekins = self.ebins + self.efermi

        self.dataprocessing(norm_mode, sigma, order, curvature)

    def dataprocessing(
        self,
        norm_mode='no',
        sigma=None,
        order=2,
        curvature=None
    ):
        """Use data post-processing procedure.

        Args:
            norm_mode (string, optional, default='no'): if 'no', no
                normalization; if 'each', every Slice is normalized by the
                integral intensity of each Slice over their 90% of axis range;
                if 'all', the overall z values are normalize to its min and
                max;
            sigma (array, optional, default=None): Standard deviation for
                Gaussian kernel. The standard deviations of the Gaussian filter
                are given for each axis as a sequence, or as a single number,
                in which case it is equal for all axes;
            order (integer, optional, default=2): The order of the filter along
                each axis is given as a sequence of integers, or as a single
                number. An order of 0 corresponds to convolution with a
                Gaussian kernel. A positive order corresponds to convolution
                with that derivative of a Gaussian. For example, an order of 2
                corresponds to convolution with that second derivative of a
                Gaussian;
            curvature (integer, optional, default=None): if not None the data
                are the curvature of the signal.
        """
        self.data = np.copy(self.data_init)

        if norm_mode == 'each':
            self.data = navplt.norm(self.data, mode=norm_mode, axis=1)

        if sigma is not None:
            self.data = postprocessing(self.data, sigma, order, curvature)

        if norm_mode == 'all':
            self.data = navplt.norm(self.data, mode=norm_mode, axis=1)

    def get_yax(self, yname='auto'):
        """Return the selected y axis.

        Args:
            yname (string, optional, auto/eef/ekin): Select the y variable.
        Returns:
            yax, yname
        """
        if yname == 'auto':
            yname = 'eef'
            yax = self.ebins
        elif yname == 'ekin':
            if hasattr(self, 'ekin'):
                yax = self.ekins
            else:
                print('Attribute ekin does not exist, using ebin')
                yax = self.ebins
                yname = 'eef'
        elif yname == 'eef':
            yax = self.ebins

        return yax, yname

    def path_show(
            self, ax=None, lc='k', textcolor='r', textsize=14, xytext=(8, 5)):
        """Plot the path.

        Args:
            ax (matplotlib.axes, optional): Axes, if None it is created inside
                the function;
            lc (string, optional): line color;
            textcolor (string, optional): text color of the labels;
            textsize (string, optional): text size of the labels;
            xytext (array, optional): relative text position of the labels.

        """
        if ax is None:
            fig, ax = plt.subplots(1)

        ax.scatter(self.kx_pts, self.ky_pts, c='k', s=12, zorder=5)
        ax.quiver(
            self.kx_pts[:-1],
            self.ky_pts[:-1],
            (self.kx_pts[1:] - self.kx_pts[:-1]),
            (self.ky_pts[1:] - self.ky_pts[:-1]),
            scale_units='xy',
            angles='xy',
            scale=1,
            fc=lc,
        )

        if self.klabels is not None:
            for klab, xlab, ylab in zip(
                self.klabels, self.kx_pts, self.ky_pts
            ):
                ax.annotate(
                    klab,
                    (xlab, ylab),
                    xycoords='data',
                    xytext=xytext,
                    textcoords='offset points',
                    color=textcolor,
                    size=textsize,
                    ha="center",
                    va="center"
                )

    def show(
        self,
        yname='auto',
        cmap='magma_r',
        ax=None,
        z_range=None,
        style=None,
        print_out=False,
        cmapscale="linear",
        vln=True,
        lc='k',
        ls='-',
        lw=1
    ):
        """Plot the isomap.

        Args:
            yname (string, optional, auto/eef/ekin): Select the y variable;
            cmap (string, optional): Colormap;
            ax (matplotlib.axes, optional): Axes, if None it is created inside
                the function;
            z_range (array, optional, [min, max]): Extreme values for z scale;
            style (string, optional): See navplt.give_style function;
            print_out (boolean, optional): If True, print the axes dimension
                with them adaptations;
            cmapscale (string, optional, linear/log/power): Select the scale
                mode.
        Returns:
            The returned value is the object matplotlib.collections.QuadMesh
                from ax.pcolormesh.

        """
        xname = 'kx'
        xax = self.krho

        yax, yname = self.get_yax(yname)

        if style is None:
            style = xname + '_' + yname

        qmisok = navplt.pimage(
            xax,
            yax,
            self.data,
            cmap=cmap,
            ax=ax,
            z_range=z_range,
            style=style,
            print_out=print_out,
            cmapscale=cmapscale
        )

        ax = qmisok.axes
        ax.set_xlim(0, self.krho.max())

        if vln:
            v_lines = self.krho[self.k_pts_bin]
            ax.set_xticks(v_lines)
            if self.klabels is not None:
                ax.set_xticklabels(self.klabels)
            else:
                ax.set_xticklabels('')

            for v_line in v_lines[1:-1]:
                ax.axvline(v_line, color=lc, linestyle=ls, linewidth=lw)

            ax.set_xlabel('')
        else:
            ax.set_xlabel(r'$Momentum$ ($\AA^{-1}$)')

        return qmisok

    def export_as_nxs(
        self,
        file_path,
        yname='auto',
        title='NavARP_isok'
    ):
        """Save the isomap as NXdata nexus class.

        Args:
            yname (string, optional, auto/eef/ekin): Select the y variable;
            title (string, optiona): data title.
        """
        yax, yname = self.get_yax(yname)

        f = h5py.File(file_path, "w")  # create the HDF5 NeXus file
        f.attrs[u"default"] = u"isok"

        nxdata = f.create_group(u"isok")
        nxdata.attrs[u"NX_class"] = u"NXdata"

        nxdata.create_dataset(u"title", data=title)

        axes_names = ['', 'momentum']
        xdata = nxdata.create_dataset(axes_names[1], data=self.krho)
        xdata.attrs[u"units"] = u"1/angstrom"
        xdata.attrs[u"axis"] = 2
        xdata.attrs[u"primary"] = 1

        if yname == 'eef':
            axes_names[0] = "binding_energies"
        else:
            axes_names[0] = "kinetic_energies"
        ydata = nxdata.create_dataset(axes_names[0], data=yax)
        ydata.attrs[u"axis"] = 1
        ydata.attrs[u"primary"] = 1
        ydata.attrs[u"units"] = u"eV"

        data_nx = nxdata.create_dataset(
            u"data",
            data=self.data.transpose(),
            compression='gzip',
            chunks=True
        )
        data_nx.attrs[u"units"] = u"counts"
        nxdata.attrs[u"signal"] = u"data"

        nxdata.attrs[u"axes"] = [axesn.encode('utf8') for axesn in axes_names]

        f.close()

    def export_as_itx(
        self,
        file_path,
        yname='auto',
        wave_name='NavARP_isok'
    ):
        """Save the isomap as Igor Pro Text file.

        Args:
            xname (string, optional, auto/tht/kx): Select the x variable;
            yname (string, optional, auto/eef/ekin): Select the y variable;
            wave_name (string, optiona): name of the wave.
        """
        xax = self.krho
        yax, yname = self.get_yax(yname)

        xlabel = 'Momentum [1/Angstrom]'
        if yname == 'eef':
            ylabel = "Binding Energy [eV]"
        else:
            ylabel = "Kinetic Energy [eV]"

        header = (
            "IGOR\n" +
            "WAVES/N=({0[0]},{0[1]})\t".format(self.data.shape) +
            wave_name + "\n" + "BEGIN"
        )

        footer = (
            'END\n' +
            'X SetScale/I x {},{},"{}", {};\n'.format(
                xax[0], xax[-1], xlabel, wave_name) +
            'X SetScale/I y {},{},"{}", {};\n'.format(
                yax[0], yax[-1], ylabel, wave_name) +
            'X SetScale d 0,0,"Counts [arb. units]", {};\n'.format(wave_name) +
            'X Note {}, "Isok generated by NavARP"'.format(wave_name)
        )

        np.savetxt(
            file_path,
            self.data,
            fmt='%.6f',
            delimiter=' ',
            newline='\n',
            header=header,
            footer=footer,
            comments='',
            encoding=None
        )


def postprocessing(data, sigma, order, curvature):
    """Post-processing procedure on data."""
    if order == 0:
        return gaussian_filter(data, sigma, order)
    elif order == 1:
        return data**3/gaussian_gradient_magnitude(data, sigma)
    elif order == 2:
        if isinstance(sigma, (list, tuple, np.ndarray)):
            return gaussian_laplace(data, sigma)
        else:
            if curvature is not None:
                dev1pwr2 = gaussian_filter(data, sigma, order=1)**2
                dev2 = gaussian_filter(data, sigma, order=2)
                curv_max = curvature*abs(
                    np.nanmax(dev1pwr2) - np.nanmin(dev1pwr2))
                return dev2*np.power((curv_max + dev1pwr2), -1.5)
            else:
                return gaussian_filter(data, sigma, order)



#!/usr/bin/env python

##############################################################################
##
# This file is part of NavARP
##
# Copyright 2016 CELLS / ALBA Synchrotron, Cerdanyola del Vallès, Spain
##
# NavARP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# NavARP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
##
# You should have received a copy of the GNU General Public License
# along with NavARP.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python NavARP library. It defines the class
 for the DFT data loaded from a PROCAR."""

__author__ = ["Federico Bisti"]
__license__ = "GPL"
__date__ = "21/03/2024"

import os
import re
import numpy as np

from dataclasses import dataclass

@dataclass
class NavDFTEntry:
    """NavDFTEntry is the class for the theoretical DFT band structure.

    NavDFTEntry is defined from the VASP PROCAR.

    Attributes:
        folder_path (str): Folder path of the input file;
        ebins (ndarray): Binding energy axis;
        kxyz (ndarray): k-space vectors;
        porbs (ndarray): projections on atomic orbitals;
        pspins (ndarray): projections on spins.

    """
    folder_path: str
    ebins: np.ndarray
    kxyz: np.ndarray
    porbs: np.ndarray
    pspins: np.ndarray
    info: str

    def get_info(self):
        print(self.info)


def load_npz(
    npz_file_path
):
    print("Loading npz-file")
    data = np.load(npz_file_path)
    print("npz-file loaded")
    return NavDFTEntry(
        npz_file_path,
        data['ebins'],
        data['kxyz'],
        data['porbs'],
        data['pspins'],
        data['info']
    )


def load_procar(
    folder_path,
    procar='PROCAR',
    outcar='OUTCAR',
    outcar_scf='OUTCAR_scf',
    npz_file_name="navdftentry.npz",
    create_npz=True,
    overwrite=False
):
    """Load vasp DFT band structure."""
    npz_file_path = os.path.join(folder_path, npz_file_name)

    if (not overwrite) and (os.path.isfile(npz_file_path)):
        return load_npz(npz_file_path)

    print("writing new npz-file")

    # Reading Fermi level (efermi) from OUTCAR_scf
    rege_float = r"[-+]?\d*\.?\d+"
    with open(os.path.join(folder_path, outcar_scf)) as foutscf:
        for line in foutscf:
            if 'E-fermi' in line:
                efermi = float(
                    re.search(
                        r'E-fermi\s*?[:\s]\s*{}'.format(rege_float),
                        line
                    ).group().split(":")[1])
                print('efermi = {}'.format(efermi))

    # Reading k-points (kxyz in units of 1/SCALE) from OUTCAR
    with open(os.path.join(folder_path, outcar)) as foutcar:
        for line in foutcar:
            if 'LNONCOLLINEAR' in line:
                if 'T' in line:
                    lnoncollinear = True
                else:
                    lnoncollinear = False

            elif 'LORBIT' in line:
                lorbit = int(
                    re.search(
                        r'LORBIT\s*?=\s*\d*',
                        line
                    ).group().split("=")[1]
                )

            elif 'NKPTS' in line:
                nkpts = int(
                    re.search(
                        r'NKPTS\s*?=\s*\d*',
                        line
                    ).group().split("=")[1]
                )
                nbands = int(
                    re.search(
                        r'NBANDS\s*?=\s*\d*',
                        line
                    ).group().split("=")[1]
                )

            elif r'k-points in units of 2pi/SCALE' in line:
                break

        kxyz = np.zeros((nkpts, 3))
        for ik in range(0, nkpts):
            line = foutcar.readline().split()
            kxyz[ik, 0] = float(line[0])
            kxyz[ik, 1] = float(line[1])
            kxyz[ik, 2] = float(line[2])
        kxyz *= 2*np.pi

    print("From OUTCAR")
    print(
        f"nkpts = {nkpts}; lnoncollinear = {lnoncollinear}; lorbit = {lorbit}")

    # START reading PROCAR
    fprocar = open(os.path.join(folder_path, procar), "r")
    if "_dn" in npz_file_name:
        fprocar.readline()
        fprocar.readline()
        fprocar.readline()
        while True:
            line = fprocar.readline()
            if "# of k-points:" in line:
                line = line.replace(':', ' ').split()
                print("Spin down")
                break
    else:
        fprocar.readline()
        line = fprocar.readline().replace(':', ' ').split()
        print(line)
        print("Spin up")

    nkpts = int(line[3])
    nbands = int(line[7])
    nions = int(line[11])
    print("From POSCAR")
    print(f"nkpts = {nkpts}; nbands = {nbands}; nions = {nions};")

    ebins = np.zeros((nkpts, nbands))
    ion = np.zeros(nions, dtype=int)

    is_porbs_pspins_defined = False

    for ikpt in range(nkpts):
        step = round((ikpt+1) / nkpts*100, 4)
        if not (step % 5):
            print(f"k-point = {ikpt+1}/{nkpts}")

        for band in range(nbands):
            # print(f"band = {band}")
            for line in fprocar:
                if 'band' in line:
                    break
            ebins[ikpt, band] = float(line.split()[4])
            fprocar.readline()
            info_line = fprocar.readline()
            if not is_porbs_pspins_defined:
                norbitals = len(info_line.split()) - 2
                info = ' '.join(
                    [f'{ind}:{str},' for ind, str in enumerate(
                        info_line.strip('ion').strip('tot\n').split())
                    ]
                )[:-1]

                print(f"norbitals = {norbitals}")
                if lnoncollinear:
                    nspins = 4  # sx, sy, sz, tot
                else:
                    nspins = 1  # tot
                pspins = np.zeros((nkpts, nbands, nions, norbitals+1, nspins))

                # s     py     pz     px    dxy    dyz    dz2    dxz    dx2
                if lorbit in [2, 12, 13, 14]:
                    porbs = np.zeros(
                        (nkpts, nbands, nions, norbitals), dtype=complex)
                else:
                    porbs = np.array([])
                is_porbs_pspins_defined = True

            for ispin in range(nspins):
                # print(f"ispin = {ispin}")
                for ion in range(0, nions):
                    # print(f"ion = {ion}")
                    pspins[ikpt, band, ion, :, ispin] = [
                        float(num) for num in fprocar.readline().split()[1:]]
                if nions != 1:
                    fprocar.readline()

            if lorbit in [2, 12, 13, 14]:
                for line in fprocar:
                    if 'ion' in line:
                        break

                appoR = []
                appoI = []
                for ion in range(0, nions):
                    line = fprocar.readline().split()
                    nums = [
                        float(num) for num in line[1:]]
                    if len(nums) >= norbitals*2:
                        appoR.append(nums[0:norbitals*2:2])
                        appoI.append(nums[1:norbitals*2:2])
                    else:
                        appoR.append(nums)
                        appoI.append([
                        float(num) for num in fprocar.readline().split()[1:]])
                porbs[ikpt, band, :, :] = np.array(appoR) + np.array(appoI)*1j

    ebins -= efermi
    if create_npz:
        np.savez(
            npz_file_path,
            kxyz=kxyz,
            ebins=ebins,
            porbs=porbs,
            pspins=pspins,
            info=info
        )
    fprocar.close()

    return NavDFTEntry(
        os.path.join(folder_path, procar), ebins, kxyz, porbs, pspins, info)


def get_efermi_from_outcar(
    outcar_path='OUTCAR',
    print_efermi='False'
):
    # Reading Fermi level (efermi) from OUTCAR_scf
    rege_float = r"[-+]?\d*\.?\d+"
    with open(outcar_path) as foutscf:
        for line in foutscf:
            if 'E-fermi' in line:
                efermi = float(
                    re.search(
                        r'E-fermi\s*?[:\s]\s*{}'.format(rege_float),
                        line
                    ).group().split(":")[1])
                if print_efermi:
                    print('efermi = {}'.format(efermi))
    return efermi



#!/usr/bin/env python

##############################################################################
##
# This file is part of NavARP
##
# Copyright 2016 CELLS / ALBA Synchrotron, Cerdanyola del Vallès, Spain
##
# NavARP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# NavARP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
##
# You should have received a copy of the GNU General Public License
# along with NavARP.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python NavARP library. It defines the functions
for transforming the angular coordinates into the k-space."""

__author__ = ["Federico Bisti"]
__license__ = "GPL"
__date__ = "08/06/2017"

import numpy as np
# from numba import jit                       # speed up the code

from scipy.optimize import newton_krylov


# @jit
def asarray(variable):
    """Transform variable as ndarray

    Args:
        variable (float, int, list, ndarray): variable to transform
    Returns:
        variable (ndarray): variable transformed
    """
    if not isinstance(variable, np.ndarray):  # if not ndarray, def from it
        if isinstance(variable, list):
            variable = np.array(variable)
        else:
            variable = np.array([variable])
    elif variable.shape == ():  # if scalar ndarray, def a array ndarray
        variable = np.array([variable])
    return variable


def get_p_hv_along_slit(hv, tht_ph):
    """Calculates photon momentum along the slit,
        i.e. the intersection between the plane of analyzer axis and slit,
        and the plane of the sample surface.

    Args:
        hv (ndarray): Photon energy;
        tht_ph (ndarray): tht angle between surface normal and photons.
    Returns:
        p_hv_along_slit (ndarray): Photon momentum in plane along theta.
    """
    hv = asarray(hv)
    tht_ph = asarray(tht_ph)

    const = 2*np.pi/12400
    kx_hv = const*hv[:, None]*np.sin(np.radians(tht_ph[None, :]))

    return kx_hv


def get_p_hv_perp_slit(hv, tht_ph, phi_ph):
    """Calculates photon momentum perpendicular to the slit,
        i.e. the intersection between plane of analyzer axis and the
        perpendicular to the slit, and the plane of the sample surface.

    Args:
        hv (ndarray): Photon energy;
        tht_ph (ndarray): tht angle between surface normal and photons.
        phi_ph (ndarray): phi angle between surface normal and photons.
    Returns:
        p_hv_along_slit (ndarray): Photon momentum in plane along phi.
    """
    hv = asarray(hv)
    phi_ph = asarray(phi_ph)
    tht_ph = asarray(tht_ph)

    if len(hv.shape) > 1:
        print("Value Error: hv cannot have a dimension > 1.")
        return None

    const = 2*np.pi/12400
    ky_hv = const*hv*(np.cos(np.radians(tht_ph[None, :])) *
                      np.sin(np.radians(phi_ph[:, None])))

    return ky_hv


def get_p_hv_perp_sample(hv, tht_ph, phi_ph):
    """Calculates photon momentum perpendicular to the sample surface

    Args:
        hv (ndarray): Photon energy;
        tht_ph (ndarray): tht angle between surface normal and photons.
        phi_ph (ndarray): phi angle between surface normal and photons.
    Returns:
        p_hv_perp_sample (ndarray): Photon momentum in plane along theta.
    """
    hv = asarray(hv)
    tht_ph = asarray(tht_ph)
    phi_ph = asarray(phi_ph)

    if len(phi_ph.shape) > 1:
        print("Value Error: phi_ph cannot have a dimension > 1.")
        return None

    const = 2*np.pi/12400
    kz_hv = -const*hv[:, None]*(np.cos(np.radians(tht_ph[None, :])) *
                                np.cos(np.radians(phi_ph)))
    return kz_hv


def get_k_along_slit(e_kins, tht, tht_an, p_hv=False, hv=None, tht_ap=None):
    """Calculates k vector along the slit, i.e. the intersection between the
        plane of analyzer axis and slit, and the plane of the sample surface.

    Args:
        e_kins (ndarray): Kinetic energies (usually it is the energy axis of
            the analyzer data);
        tht (ndarray): Angular axis of the analyzer data;
        tht_an (float): angle between analyzer axis (a) and normal (n) to the
            surface along the plane of the slit (theta, so the name "tht_an");
        p_hv (boolean, optional): If True, add photon momentum;
        hv (ndarray): Photon energies;
        tht_ap (float): angle between analyzer axis (a) and photons (p) along
            the plane of the slit (theta, so the name "tht_ap").
    Returns:
        kx (ndarray): k vector along the slit
    """
    # Transform possible input (ndarray, list or value) into ndarray
    e_kins = asarray(e_kins)
    tht = asarray(tht)
    hv = asarray(hv)

    # Angle from normal of emitted electrons in radians
    tht_en = np.radians(tht-tht_an)

    # Calculate the electron momentum without photon one
    # "None" is for broadcasting
    # kx shape is (e_kins.shape[0], tht_en.shape[0], e_kins.shape[1]) or
    # (tht_en.shape[0], e_kins.shape[0])
    C_0 = 0.5124
    if len(e_kins.shape) == 2:
        kx = C_0*np.sqrt(e_kins[:, None, :])*np.sin(tht_en[None, :, None])
    else:
        kx = C_0*np.sqrt(e_kins[None, :])*np.sin(tht_en[:, None])

    # Substract the photon momentum if p_hv=True
    if p_hv:
        kx_hv = get_p_hv_along_slit(hv, -tht_ap + tht_an)
        if len(kx.shape) == 3:
            kx -= kx_hv[:, :, None]
        elif kx.shape[1] == kx_hv.shape[0]:
            kx -= kx_hv[None, :, 0]
        else:
            kx -= kx_hv

    kx = np.squeeze(kx)
    return kx


def get_k_perp_slit(e_kins, tht, tht_an, phi, phi_an,
                    p_hv=False, hv=None, tht_ap=None, phi_ap=None):
    """Calculates k vector perpendicular to the slit, i.e. the intersection
        between plane of analyzer axis and the perpendicular to the slit, and
        the plane of the sample surface.

    Args:
        e_kins (ndarray): Kinetic energies (usually it is the energy axis of
            the analyzer data);
        tht (ndarray): Angular axis of the analyzer data;
        tht_an (float): angle between analyzer axis (a) and normal (n) to the
            surface along the plane of the slit (theta, so the name "tht_an");
        phi (ndarray): angle perpendicular to the slit, this is the axis of
            the scan in the case of the deflector;
        phi_an (float): angle between analyzer axis (a) and normal (n) along
            the plane perpendicular to the slit (phi, so the name "phi_an"),
            this is the axis of the scan in the case of tilt rotation;
        p_hv (boolean, optional): If True, add photon momentum.
        hv (ndarray): Photon energies;
        tht_ap (float): angle between analyzer axis (a) and photons (p) along
            the plane of the slit (theta, so the name "tht_ap");
        phi_ap (float): angle between analyzer axis (a) and photons (p) along
            the plane perpendicular to the slit (phi, so the name "phi_ap");
    Returns:
        ky (ndarray): k vector perpendicular to the slit
    """
    # Transforming possible input (ndarray, list or value) into ndarray
    e_kins = asarray(e_kins)
    tht = asarray(tht)
    phi = asarray(phi)
    phi_an = asarray(phi_an)

    # Angles from normal of emitted electrons in radians
    phi_en = np.radians(phi-phi_an)
    tht_en = np.radians(tht-tht_an)

    if len(e_kins.shape) > 1:
        print("Value Error: e_kins cannot have a dimension > 1.")
        return None
    if len(phi.shape) > 1 and len(phi_an.shape) > 1:
        print("Value Error: both phi and phi_an cannot have a dimension > 1.")
        return None

    # Calculate the electron momentum without photon one
    # "None" is for broadcasting
    # ky shape is (phi_en.shape[0], tht_en.shape[0], e_kins.shape[0])
    C_0 = 0.5124
    if len(phi_en.ravel()) > 1:
        ky = C_0*np.sqrt(e_kins[None, None, :])*(
                np.cos(tht_en[None, :, None])*np.sin(phi_en[:, None, None]))
    else:
        ky = C_0*np.sqrt(e_kins[None, :])*(
                np.cos(tht_en[:, None])*np.sin(phi_en[0]))

    # Substract the photon momentum if p_hv=True
    if p_hv:
        ky_hv = get_p_hv_perp_slit(hv, -tht_ap+tht_an, -phi_ap+phi_an)
        if len(phi_en.ravel()) > 1:
            ky -= ky_hv[:, None]
        else:
            ky -= ky_hv

    ky = np.squeeze(ky)
    return ky


def get_k_perp_sample(e_kins, inn_pot, k_along_slit, k_perp_slit,
                      p_hv=False, hv=None, tht_ap=None, phi_ap=None,
                      tht_an=None, phi_an=None):
    """Calculates k vector perpendicular to sample surface.

    Args:
        e_kins (ndarray): Kinetic energies (usually it is the energy axis of
            the analyzer data);
        inn_pot (float): Inner potential, corresponding to the energy of the
            bottom of the valence band referenced to vacuum level;
        k_along_slit (ndarray): k vector along the slit
        k_perp_slit (ndarray): k vector perpendicular to the slit
        p_hv (boolean, optional): If True, add photon momentum.
        hv (ndarray): Photon energies;
        tht_ap (float): angle between analyzer axis (a) and photons (p) along
            the plane of the slit (theta, so the name "tht_ap");
        phi_ap (float): angle between analyzer axis (a) and photons (p) along
            the plane perpendicular to the slit (phi, so the name "phi_ap");
        tht_an (float): angle between analyzer axis (a) and normal (n) to the
            surface along the plane of the slit (theta, so the name "tht_an");
        phi_an (float): angle between analyzer axis (a) and normal (n) along
            the plane perpendicular to the slit (phi, so the name "phi_an"),
            this is the axis of the scan in the case of tilt rotation.
    Returns:
        kz (ndarray): k vector perpendicular to sample surface.
    """
    # Transforming possible input (ndarray, list or value) into ndarray
    tht_an = asarray(tht_an)
    phi_an = asarray(phi_an)
    e_kins = asarray(e_kins)
    k_along_slit = asarray(k_along_slit)
    k_perp_slit = asarray(k_perp_slit)

    if len(phi_an.shape) > 1:
        print("Value Error: phi_an cannot have a dimension > 1.")
        return None
    if len(k_perp_slit.shape) > 1:
        print("Value Error: phi_an cannot have a dimension > 1.")
        return None

    ekin_xy = (np.power(k_along_slit, 2) + np.power(k_perp_slit, 2))

    C_2 = 0.5124**2
    if len(e_kins.shape) == 2:
        kz = np.sqrt(C_2*(e_kins[:, None, :] + inn_pot) - ekin_xy)
    else:
        kz = np.sqrt(C_2*(e_kins[None, :] + inn_pot) - ekin_xy)

    # Substract the photon momentum if p_hv=True
    if p_hv:
        kz_hv = get_p_hv_perp_sample(hv, -tht_ap+tht_an, -phi_ap+phi_an)
        if len(kz.shape) == 3:
            kz -= kz_hv[:, :, None]
        elif kz.shape[1] == kz_hv.shape[0]:
            kz -= kz_hv[None, :, 0]
        else:
            kz -= kz_hv

    kz = np.squeeze(kz)
    return kz


def get_tht_an(e_kin_p, tht_p, k_along_slit_p, tht_an_init=0,
               p_hv=False, hv_p=None, tht_ap=None):
    """Calculates angle between analyzer axis (a) and normal (n) to the
            surface along the plane of the slit (theta, so the name "tht_an").

    Args:
        e_kin_p (float): Kinetic energy value of the reference point;
        tht_p (float): Angular value of the reference point along the angular
            axis of the analyzer data;
        k_along_slit_p (float): k vector value of the reference point, along
            the slit direction;
        tht_an_init (float, optional): Initial guess for the tht_an;
        p_hv (boolean, optional): If True, add photon momentum;
        hv_p (float, optional): Photon energy for the reference point,
            requested if p_hv==True;
        tht_ap (float, optional): angle between analyzer axis (a) and
            photons (p) along the plane of the slit (theta, so the name
            "tht_ap"), requested if p_hv==True.
    Returns:
        tht_an (float): angle between analyzer axis (a) and normal (n) to the
            surface along the plane of the slit (theta, so the name "tht_an").
    """
    return newton_krylov(
        lambda tht_an_x: get_k_along_slit(
            e_kin_p,
            tht_p,
            tht_an_x,
            p_hv=p_hv,
            hv=hv_p,
            tht_ap=tht_ap
        ) - k_along_slit_p,
        tht_an_init
    )


def get_phi_an(e_kin_p, tht_p, tht_an, phi_p, k_perp_slit_p, phi_an_init=0,
               p_hv=False, hv_p=None, tht_ap=None, phi_ap=None):
    """Calculates angle between analyzer axis (a) and normal (n) along
            the plane perpendicular to the slit (phi, so the name "phi_an"),
            this is the axis of the scan in the case of tilt rotation;

    Args:
        e_kin_p (float): Kinetic energy value of the reference point;
        tht_p (float): Angular value of the reference point along the angular
            axis of the analyzer data;
        tht_an (float): angle between analyzer axis (a) and normal (n) to the
            surface along the plane of the slit (theta, so the name "tht_an");
        phi_p (float): Angular value of the reference point along the angle
            perpendicular to the slit;
        k_perp_slit_p (float): k vector value of the reference point,
            perpendicular to the slit;
        phi_an_init (float, optional): Initial guess for the phi_an;
        p_hv (boolean, optional): If True, add photon momentum;
        hv_p (float, optional): Photon energy for the reference point,
            requested if p_hv==True;
        tht_ap (float, optional): angle between analyzer axis (a) and
            photons (p) along the plane of the slit (theta, so the name
            "tht_ap"), requested if p_hv==True;
        phi_ap (float): angle between analyzer axis (a) and photons (p) along
            the plane perpendicular to the slit (phi, so the name "phi_ap");
    Returns:
        phi_an (float): angle between analyzer axis (a) and normal (n) along
            the plane perpendicular to the slit (phi, so the name "phi_an"),
            this is the axis of the scan in the case of tilt rotation.
    """
    return newton_krylov(
        lambda phi_an_x: get_k_perp_slit(
            e_kin_p,
            tht_p,
            tht_an,
            phi_p,
            phi_an_x,
            p_hv=p_hv,
            hv=hv_p,
            tht_ap=tht_ap,
            phi_ap=phi_ap
        ) - k_perp_slit_p,
        phi_an_init
    )


def get_hv_from_kxyz(e_bin_p, work_fun, inn_pot,
                     k_along_slit_p, k_perp_slit_p, k_perp_sample_p,
                     hv_p_init=30,
                     p_hv=False, tht_ap=0, phi_ap=0,
                     tht_an=None, phi_an=None):
    """Calculates photon energy of a point from k-space coordinates

    Args:
        e_bin_p (float): Binding energy value of the point;
        work_fun (float): analyzer work function;
        inn_pot (float): Inner potential, corresponding to the energy of the
            bottom of the valence band referenced to vacuum level;
        k_along_slit_p (float): k vector value of the reference point, along
            the slit direction;
        k_perp_slit_p (float): k vector value of the reference point,
            perpendicular to the slit;
        k_perp_sample_p (float): k vector perpendicular to sample surface
        hv_p_init (float): Initial guess for the photon energy of the
            reference point;
        p_hv (boolean, optional): If True, add photon momentum;
        tht_ap (float, optional): angle between analyzer axis (a) and
            photons (p) along the plane of the slit (theta, so the name
            "tht_ap"), requested if p_hv==True;
        phi_ap (float): angle between analyzer axis (a) and photons (p) along
            the plane perpendicular to the slit (phi, so the name "phi_ap");
        tht_an (float): angle between analyzer axis (a) and normal (n) to the
            surface along the plane of the slit (theta, so the name "tht_an");
        phi_an (float): angle between analyzer axis (a) and normal (n) along
            the plane perpendicular to the slit (phi, so the name "phi_an"),
            this is the axis of the scan in the case of tilt rotation.
    Returns:
        hv_p (float): Photon energy of the point;
    """
    return newton_krylov(
        lambda x_hv_p: get_k_perp_sample(
            (x_hv_p - work_fun + e_bin_p),
            inn_pot,
            k_along_slit_p,
            k_perp_slit_p,
            p_hv=p_hv,
            hv=x_hv_p,
            tht_ap=tht_ap,
            phi_ap=phi_ap,
            tht_an=tht_an,
            phi_an=phi_an
        ) - k_perp_sample_p,
        hv_p_init
    )


def get_inn_pot_from_kxyz(e_kin_p,
                          k_along_slit_p, k_perp_slit_p, k_perp_sample_p,
                          inn_pot_init=10,
                          p_hv=False, hv_p=None, tht_ap=0, phi_ap=0,
                          tht_an=None, phi_an=None):
    """Calculates photon energy of a point from k-space coordinates

    Args:
        e_kin_p (float): Kinetic energy value of the reference point;
        inn_pot_init (float): Initial guess for the inner potential,
            corresponding to the energy of the bottom of the valence band
            referenced to vacuum level;
        k_along_slit_p (float): k vector value of the reference point, along
            the slit direction;
        k_perp_slit_p (float): k vector value of the reference point,
            perpendicular to the slit;
        k_perp_sample_p (float): k vector perpendicular to sample surface
        hv (ndarray): Photon energies;
        p_hv (boolean, optional): If True, add photon momentum;
        hv_p (float, optional): Photon energy for the reference point,
            requested if p_hv==True;
        tht_ap (float, optional): angle between analyzer axis (a) and
            photons (p) along the plane of the slit (theta, so the name
            "tht_ap"), requested if p_hv==True;
        phi_ap (float): angle between analyzer axis (a) and photons (p) along
            the plane perpendicular to the slit (phi, so the name "phi_ap");
        tht_an (float): angle between analyzer axis (a) and normal (n) to the
            surface along the plane of the slit (theta, so the name "tht_an");
        phi_an (float): angle between analyzer axis (a) and normal (n) along
            the plane perpendicular to the slit (phi, so the name "phi_an"),
            this is the axis of the scan in the case of tilt rotation.
    Returns:
        inn_pot (float): Inner potential, corresponding to the energy of the
            bottom of the valence band referenced to vacuum level.
    """
    return newton_krylov(
        lambda x_inn_pot: get_k_perp_sample(
            e_kin_p,
            x_inn_pot,
            k_along_slit_p,
            k_perp_slit_p,
            p_hv=p_hv,
            hv=hv_p,
            tht_ap=tht_ap,
            phi_ap=phi_ap,
            tht_an=tht_an,
            phi_an=phi_an
        ) - k_perp_sample_p,
        inn_pot_init
    )


def get_surface_normal(entry, refs_p):
    """Calculates tht_an, phi_an, scans_0 and phi.

    Args:
        entry (NavEntry class): the class for the data explored by NavARP;
        refs_p (dict): dictionary of the reference point in angle and k space.
    Returns:
        hv_p (float): Photon energy of the point;
    """

    tht_p = refs_p["tht_p"]
    hv_p = refs_p["hv_p"]
    e_kin_p = refs_p["e_kin_p"]
    k_along_slit_p = refs_p["k_along_slit_p"]
    p_hv = refs_p["p_hv"]

    tht_ap = entry.analyzer.tht_ap
    phi_ap = entry.analyzer.phi_ap

    # The photon momentum shifts the G point from (Kx, Ky) = (0, 0)
    # In the following code-line the real angle value is calculated
    # from the G point by removing the photon momentum contribution
    # in this way G point remains fixed for all the photon energies
    # because the photon momentum is taken into account
    tht_an = get_tht_an(e_kin_p, tht_p, k_along_slit_p, tht_an_init=tht_p,
                        p_hv=p_hv, hv_p=hv_p, tht_ap=tht_ap)
    print("tht_an = {:0.3f}".format(tht_an))

    if entry.scan_type == "hv":
        phi_an = 0
        scans_0 = 0
    elif (entry.scan_type == "polar" or entry.scan_type == "tilt"):
        tilt_p = refs_p["scan_p"]
        phi_p = entry.defl_angles
        phi_an_p = get_phi_an(e_kin_p, tht_p, tht_an, phi_p, refs_p["ks_p"],
                              phi_an_init=0,
                              p_hv=p_hv, hv_p=hv_p,
                              tht_ap=tht_ap, phi_ap=phi_ap)
        print("phi_an_p = {:0.3f}".format(phi_an_p))
        scans_0 = tilt_p - phi_an_p
        print("scans_0 = {:0.3f}".format(scans_0))
        phi_an = entry.scans - scans_0
    elif entry.scan_type == "deflector":
        phi_p = refs_p["scan_p"]
        phi_an_p = get_phi_an(e_kin_p, tht_p, tht_an, phi_p, refs_p["ks_p"],
                              phi_an_init=0,
                              p_hv=p_hv, hv_p=hv_p,
                              tht_ap=tht_ap, phi_ap=phi_ap)
        print("phi_an_p = {:0.3f}".format(phi_an_p))
        scans_0 = phi_an_p
        print("scans_0 = {:0.3f}".format(scans_0))
        phi_an = phi_an_p
    else:
        phi_an = 0
        scans_0 = 0

    return tht_an, phi_an, scans_0


def get_k_isoen(entry, e_kin_val):
    """Calculates k vectors of the iso-energy cut.

    Args:
        entry (NavEntry class): the class for the data explored by NavARP;
        e_kin_val (float or ndarray): Kinetic energy/ies of iso-energy cut.
    """

    kx_isoen = get_k_along_slit(
        e_kins=e_kin_val,
        tht=entry.angles,
        tht_an=entry.tht_an,
        p_hv=entry.p_hv,
        hv=entry.hv,
        tht_ap=entry.analyzer.tht_ap
    )

    if entry.scan_type == "hv":
        ks_isoen = get_k_perp_sample(
            e_kins=e_kin_val,
            inn_pot=entry.inn_pot,
            k_along_slit=kx_isoen,
            k_perp_slit=entry.k_perp_slit_for_kz,
            p_hv=entry.p_hv,
            hv=entry.hv,
            tht_ap=entry.analyzer.tht_ap,
            phi_ap=entry.analyzer.phi_ap,
            tht_an=entry.tht_an,
            phi_an=entry.phi_an
        )
    elif (entry.scan_type == "polar" or entry.scan_type == "tilt"):
        ks_isoen = get_k_perp_slit(
            e_kins=e_kin_val,
            tht=entry.angles,
            tht_an=entry.tht_an,
            phi=entry.defl_angles,
            phi_an=entry.phi_an,
            p_hv=entry.p_hv,
            hv=entry.hv,
            tht_ap=entry.analyzer.tht_ap,
            phi_ap=entry.analyzer.phi_ap
        )
    elif entry.scan_type == "deflector":
        ks_isoen = get_k_perp_slit(
            e_kins=e_kin_val,
            tht=entry.angles,
            tht_an=entry.tht_an,
            phi=entry.scans,
            phi_an=entry.phi_an,
            p_hv=entry.p_hv,
            hv=entry.hv,
            tht_ap=entry.analyzer.tht_ap,
            phi_ap=entry.analyzer.phi_ap
        )
    elif entry.scan_type == "azimuth":
        scans_rad = np.radians(entry.scans-entry.scans_0)
        krho = kx_isoen
        kx_isoen = np.squeeze(krho * np.cos(scans_rad)[:, None, None])
        ks_isoen = np.squeeze(krho * np.sin(scans_rad)[:, None, None])
    else:
        ks_isoen = entry.scans

    return kx_isoen, ks_isoen



# -*- coding: utf-8 -*-
"""
Created on Fri Jul 12 10:35:38 2019

@author: fbisti
"""

import sys
from PyQt5 import QtWidgets

app = QtWidgets.QApplication(sys.argv)

screen = app.primaryScreen()
print('Screen: %s' % screen.name())
size = screen.size()
print('Size: %d x %d' % (size.width(), size.height()))
rect = screen.availableGeometry()
print('Available: %d x %d' % (rect.width(), rect.height()))


#!/usr/bin/env python

##############################################################################
##
# This file is part of NavARP
##
# Copyright 2016 CELLS / ALBA Synchrotron, Cerdanyola del Vallès, Spain
##
# NavARP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# NavARP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
##
# You should have received a copy of the GNU General Public License
# along with NavARP.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python NavARP library. It defines the functions
for plotting the data using matplotlib.pyplot ."""

__author__ = ["Federico Bisti"]
__license__ = "GPL"
__date__ = "28/04/2017"

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as colors
try:
    import colorcet as cc  # Additional perceptually uniform cmaps
except ImportError:
    print("WARNING: colorcet package not found. " +
          "It is not necessary but, if present, "
          "additional perceptually uniform cmaps are available.\n")


def pimage(x, y, z, cmap='Greys', ax=None, z_range=None, style=None,
           print_out=False, cmapscale="linear"):
    """ The function is adapting a more generic input format for pcolormesh.

    Args:
        x (ndarray): Array for the horizontal axis
        y (ndarray): Array for the vertical axis
        z (ndarray): Intensity matrix
        cmap (string, optional): colormap
        ax (matplotlib.axes, optional): Axes, if None it is created inside the
            function
        z_range (array, optional, [min, max]): Extreme values for z scale
        style (string, optional): See give_style function
        print_out (boolean, optional): If True, print the axes dimension with
            them adaptations
    Returns:
        The returned value is the object matplotlib.collections.QuadMesh from
            ax.pcolormesh
    """

    if ax is None:
        plt.figure()
        ax = plt.subplot(111)

    if style is not None:
        give_style(ax, style)

    x_ext = extend_array(x, z, 'X')
    y_ext = extend_array(y, z, 'Y')

    if print_out:
        print("Dimensions of Z {0} are compatible with ".format(z.shape),
              "X {0} and Y {1} after reshaping into ".format(x.shape,
                                                             y.shape),
              "X_ext {0} and Y_ext {1}.".format(x_ext.shape, y_ext.shape))

    # Set color map range and scale
    if z_range is None:
        vmin = np.nanmin(z)
        vmax = np.nanmax(z)
    else:
        vmin = min(z_range)
        vmax = max(z_range)

    if cmapscale == "log":
        z_min = z.min()
        if z_min <= 0:
            z_max = z.max()
            z = (z-z_min) + (z_max-z_min)*1e-4
            vmin = (vmin-z_min) + (z_max-z_min)*1e-4
            vmax = (vmax-z_min) + (z_max-z_min)*1e-4

    norm = get_cmapscale(cmapscale, vmin=vmin, vmax=vmax, clip=False)

    # return the matplotlib.collections.QuadMesh
    return ax.pcolormesh(x_ext, y_ext, z, cmap=cmap, norm=norm)


def extend_array(array, image, array_label):
    i_array_shape = array.shape

    if 'X' in array_label:
        ref_shape = image.shape[1]
        other_shape = image.shape[0]
    elif 'Y' in array_label:
        ref_shape = image.shape[0]
        other_shape = image.shape[1]

    # If array is 1-dim array then its dimension has to be of the same size of
    # the second (first) dimension of image if it is X (Y)
    # Otherwise make it as a 2-dim matrix and extend it in the following part
    if len(array.shape) == 1:
        if array.shape[0] == ref_shape:
            if array.shape[0] == 1:
                darray = 0.5
                array_ext = np.array([array[0] - darray, array[0] + darray])
            else:
                array_ext = array + np.diff(
                        np.append(array,
                                  array[-1]+(array[-1]-array[-2])))*0.5
                array_ext = np.append(array[0]-(array[1]-array[0])*0.5,
                                      array_ext)
        elif array.shape[0] == other_shape:
            array = np.tile(array, (ref_shape, 1))
        else:
            print("Error:",
                  "Dimensions of Z {0} are incompatible".format(image.shape),
                  " with ", array_label, " {0}".format(i_array_shape))
            return

    # If array is 2-dim array then its dimension has to be of the same of image
    # or it can be transposed, check it and adjust it accordling
    if len(array.shape) == 2:
        if (array.shape[0] == image.shape[1] and
                array.shape[1] == image.shape[0]):
            array = np.transpose(array, (1, 0))
        elif not (array.shape[0] == image.shape[0] and
                  array.shape[1] == image.shape[1]):
            print("Error:",
                  "Dimensions of Z {0} are incompatible".format(image.shape),
                  " with ", array_label, " {0}".format(i_array_shape))
            return

        # add column
        if array.shape[1] == 1:
            darray = 0.5
            col_add = array[:, 0] + darray
        else:
            col_add = array[:, -1] + (array[:, -1] - array[:, -2])
        array_ext0 = array + np.diff(np.hstack((array, col_add[:, None])))*0.5
        if array.shape[1] == 1:
            col_add = array[:, 0] - darray*0.5
        else:
            col_add = array[:, 0] - (array[:, 1] - array[:, 0])*0.5
        array_ext0 = np.hstack((col_add[:, None], array_ext0))

        # add raw
        if array.shape[0] == 1:
            darray = 0.5
            raw_add = array_ext0[0, :] + darray
        else:
            raw_add = array_ext0[-1, :] + (array_ext0[-1, :] -
                                           array_ext0[-2, :])
        array_ext = array_ext0 + np.diff(
                np.vstack((array_ext0, raw_add[None, :])), axis=0)*0.5
        if array.shape[0] == 1:
            darray = 0.5
            raw_add = array_ext0[0, :] - darray*0.5
        else:
            raw_add = array_ext0[0, :]-(array_ext0[1, :]-array_ext0[0, :])*0.5
        array_ext = np.vstack((raw_add[None, :], array_ext))

    if len(array.shape) > 2:
        print("Error:",
              "Dimensions of Z {0} are incompatible".format(image.shape),
              " with ", array_label, " {0}".format(i_array_shape))
        return

    return array_ext


def get_cmapscale(cmapscale="linear", vmin=None, vmax=None, clip=False):
    """ The function generate linear/log/power colormap scale.

    Args:
        cmapscale (string, optional, linear/log/power): Select the scale mode.
        vmin, vmax (scalar, optional): Colorscale min and max.
            If None, suitable min/max values are automatically chosen by the
            Normalize instance.
        clip (boolean, optional): if is True, masked values are set to 1;
            otherwise they remain masked.
    Returns:
        norm: Normalize class from matplotlib.colors.Normalize
    """
    if cmapscale == "linear":
        norm = colors.Normalize(vmin=vmin, vmax=vmax, clip=clip)
    elif cmapscale == "log":
        norm = colors.LogNorm(vmin=vmin, vmax=vmax, clip=clip)
    elif cmapscale == "power":
        norm = colors.PowerNorm(gamma=0.5, vmin=vmin, vmax=vmax, clip=clip)
    else:
        print("ValueError: unknown cmapscale, returned the linear one.")
        norm = colors.Normalize(vmin=vmin, vmax=vmax, clip=clip)

    return norm


def set_cmapscale(quadmesh, cmapscale, vmin=None, vmax=None, clip=False):
    """ The function assign the cmapscale to the quadmesh.

    Args:
        quadmesh: matplotlib.collections.QuadMesh from ax.pcolormesh
        cmapscale (string, optional, linear/log/power): Select the scale mode.
        vmin, vmax (scalar, optional): Colorscale min and max.
            If None, suitable min/max values are automatically chosen by the
            Normalize instance.
        clip (boolean, optional): if is True, masked values are set to 1;
            otherwise they remain masked.
    Returns:
        None
    """
    norm = get_cmapscale(cmapscale, vmin=vmin, vmax=vmax, clip=clip)
    quadmesh.set_norm(norm)


def give_style(ax, style):
    """ The function applies a style to the pcolormesh generated by pimage.

    Args:
        ax (matplotlib.axes): Axes where to impose the style.
        style (string): 'xlabel_ylabel', xlabel and ylabel in agreement with
            label dictionary (e.g. 'tht_be' or 'kx_hv' or 'ky_kz')

    Returns:
        None
    """

    label = {'tht': r'$\theta (^\circ)$',
             'phi': r'$\phi (^\circ)$',
             'be': r'Binding Energy (eV)',
             'ekin': r'Kinetic Energy (eV)',
             'eef': r"$E-E_F$ ($eV$)",
             'ekhv': r'$E_{kin} - h\nu + work \: fun.$ ($eV$)',
             'hv': r'Photon Energy (eV)',
             'kx': r'$k_x$ ($\AA^{-1}$)',
             'ky': r'$k_y$ ($\AA^{-1}$)',
             'kz': r'$k_z$ ($\AA^{-1}$)'}

    xylabel = style.lower().split('_')
    ax.set_xlabel(label[xylabel[0]])
    ax.set_ylabel(label[xylabel[1]])

    ax.get_xaxis().set_tick_params(direction='out')
    ax.get_yaxis().set_tick_params(direction='out')


def norm(z, mode="all", axis=1):
    """ Normalize z values along the axis value between [0, 1].

    Args:
        z (ndarray): Intensity matrix
        mode (string, optional, all or each): Select the normalization mode
            if "each", every Slice are normalize by by the integral intensity
            of each Slice over their 90% of aixs range
            if "all", the overall z values are normalize to its min and max
        axis (int, optional): Axis along which a sum is performed in the case
            mode == "each"
    Returns:
        z_norm (ndarray): Intensity matrix normalized
    """
    if mode == "each":
        # normalize every Slice by the integral intensity of each Slice over
        # their 90% of angular range
        perc = 0.10
        ang_norm_i = int(z.shape[axis]*perc)
        if axis == 0:
            z_norm = z[ang_norm_i:-ang_norm_i, :]
            z_norm = np.sum(z_norm, axis=axis)/(z_norm.shape[axis])
            z_norm = np.tile(z_norm, (z.shape[axis], 1))
        elif axis == 1:
            z_norm = z[:, ang_norm_i:-ang_norm_i]
            z_norm = np.sum(z_norm, axis=axis)/(z_norm.shape[axis])
            z_norm = np.transpose(np.tile(z_norm, (z.shape[axis], 1)))
        z_norm = z/z_norm
        z_norm_min = np.nanmin(z_norm)
        z_norm_max = np.nanmax(z_norm)
        z_norm = (z_norm-z_norm_min)/(z_norm_max-z_norm_min)
    elif mode == "all":
        # normalize the matrix by its maximun
        z_min = np.nanmin(z)
        z_max = np.nanmax(z)
        z_norm = (z-z_min)/(z_max-z_min)
    else:
        print('Value Error: mode "{}" not understood,'
              ' mode can be only "all" or "each".'.format(mode))
        z_norm = None
    return z_norm


def hist(z, ax=None, z_range=None, orientation='vertical'):
    """ Plot an histogram of z divited into 256 bins.

    Args:
        z (ndarray): Intensity matrix
        ax (matplotlib.axes, optional): Axes, if None it is created inside the
            function
        z_range (array, optional, [min, max]): Extreme values for z scale
        orientation (string, optional, vertical or horizontal): Histogram
            axes orientation
    Returns:
        None
    """
    if ax is None:
        plt.figure()
        ax = plt.subplot(111)

    if z_range is None:
        zmin = z.min()
        zmax = z.max()
    else:
        zmin = min(z_range)
        zmax = max(z_range)

    ax.hist(z.ravel(), bins=256, range=(zmin, zmax), orientation=orientation,
            fc='k', ec='k')

    if orientation == 'vertical':
        ax.set_xlim(zmin, zmax)
    elif orientation == 'horizontal':
        ax.set_ylim(zmin, zmax)


# ----------------------------------------------------------------------------
# -------------------- deprecated --------------------------------------------
def pimage_depr(x, y, z, cmap='Greys', ax=None, z_range=None, style=None,
                print_out=False, cmapscale="linear"):
    """ The function is adapting a more generic input format for pcolormesh.

    Args:
        x (ndarray): Array for the horizontal axis
        y (ndarray): Array for the vertical axis
        z (ndarray): Intensity matrix
        cmap (string, optional): colormap
        ax (matplotlib.axes, optional): Axes, if None it is created inside the
            function
        z_range (array, optional, [min, max]): Extreme values for z scale
        style (string, optional): See give_style function
        print_out (boolean, optional): If True, print the axes dimension with
            them adaptations
    Returns:
        The returned value is the object matplotlib.collections.QuadMesh from
            ax.pcolormesh
    """

    if ax is None:
        plt.figure()
        ax = plt.subplot(111)

    if style is not None:
        give_style(ax, style)

    # The function can modify the shape of x and y, so saving the initial ones.
    i_x_shape = x.shape
    i_y_shape = y.shape
    # If x is a vector then the 2nd dimension of z has to be of the same size
    # Otherwise it is making x a 2D matrix
    if len(x.shape) == 1:
        if x.shape[0] == z.shape[1]:
            x_ext = x + np.diff(np.append(x, x[-1]+(x[-1]-x[-2])))*0.5
            x_ext = np.append(x[0]-(x[1]-x[0])*0.5, x_ext)
        elif x.shape[0] == z.shape[0]:
            x = np.tile(x, (z.shape[1], 1))
        else:
            print("Error1:",
                  "Dimensions of Z {0} are incompatible with ".format(z.shape),
                  "X {0} and/or Y {1} ".format(i_x_shape, i_y_shape))
            return
    # If x is a matrix must have the same size of z, but can also be transposed
    if len(x.shape) == 2:
        if x.shape[0] == z.shape[1] and x.shape[1] == z.shape[0]:
            x = np.transpose(x, (1, 0))
        elif not (x.shape[0] == z.shape[0] and x.shape[1] == z.shape[1]):
            print("Error2:",
                  "Dimensions of Z {0} are incompatible with ".format(z.shape),
                  "X {0} and/or Y {1} ".format(i_x_shape, i_y_shape))
            return
        if x.shape[1] == 1:
            col_add = x[:, 0]+0.2
        else:
            col_add = x[:, -1]+(x[:, -1]-x[:, -2])
        x_ext0 = x + np.diff(np.hstack((x, col_add[:, None])))*0.5
        if x.shape[1] == 1:
            col_add = x[:, 0]-0.2*0.5
        else:
            col_add = x[:, 0]-(x[:, 1]-x[:, 0])*0.5
        x_ext0 = np.hstack((col_add[:, None], x_ext0))
        raw_add = x_ext0[-1, :]+(x_ext0[-1, :]-x_ext0[-2, :])
        x_ext = x_ext0 + np.diff(np.vstack((x_ext0, raw_add[None, :])),
                                 axis=0)*0.5
        raw_add = x_ext0[0, :]-(x_ext0[1, :]-x_ext0[0, :])*0.5
        x_ext = np.vstack((raw_add[None, :], x_ext))
    if len(x.shape) > 2:
        print("Error3:",
              "Dimensions of Z {0} are incompatible with ".format(z.shape),
              "X {0} and/or Y {1} ".format(i_x_shape, i_y_shape))
        return

    # If y is a vector then the 1st dimension of z has to be of the same size
    # Otherwise it is making y a 2D matrix
    if len(y.shape) == 1:
        if y.shape[0] == z.shape[0]:
            y_ext = y + np.diff(np.append(y, y[-1]+(y[-1]-y[-2])))*0.5
            y_ext = np.append(y[0]-(y[1]-y[0])*0.5, y_ext)
        elif y.shape[0] == z.shape[1]:
            y = np.tile(y, (z.shape[0], 1))
        else:
            print("Error4:",
                  "Dimensions of Z {0} are incompatible with ".format(z.shape),
                  "X {0} and/or Y {1} ".format(i_x_shape, i_y_shape))
            return

    # If y is a matrix must have the same size of z, but can also be transposed
    if len(y.shape) == 2:
        if y.shape[0] == z.shape[1] and y.shape[1] == z.shape[0]:
            y = np.transpose(y, (1, 0))
        elif not (y.shape[0] == y.shape[0] and y.shape[1] == y.shape[1]):
            print("Error5:"
                  "Dimensions of Z {0} are incompatible with ".format(z.shape),
                  "X {0} and/or Y {1} ".format(i_x_shape, i_y_shape))
            return
        if y.shape[1] == 1:
            col_add = y[:, 0]+0.2
        else:
            col_add = y[:, -1]+(y[:, -1]-y[:, -2])
        y_ext0 = y + np.diff(np.hstack((y, col_add[:, None])))*0.5
        if y.shape[1] == 1:
            col_add = y[:, 0]-0.2*0.5
        else:
            col_add = y[:, 0]-(y[:, 1]-y[:, 0])*0.5
        y_ext0 = np.hstack((col_add[:, None], y_ext0))
        raw_add = y_ext0[-1, :]+(y_ext0[-1, :]-y_ext0[-2, :])
        y_ext = y_ext0 + np.diff(np.vstack((y_ext0, raw_add[None, :])),
                                 axis=0)*0.5
        raw_add = y_ext0[0, :]-(y_ext0[1, :]-y_ext0[0, :])*0.5
        y_ext = np.vstack((raw_add[None, :], y_ext))

    if len(y.shape) > 2:
        print("Error6:",
              "Dimensions of Z {0} are incompatible with ".format(z.shape),
              "X {0} and/or Y {1} ".format(i_x_shape, i_y_shape))
        return

    if print_out:
        print("Dimensions of Z {0} are compatible with ".format(z.shape),
              "X {0} and Y {1} after reshaping into ".format(i_x_shape,
                                                             i_y_shape),
              "X_ext {0} and Y_ext {1}.".format(x_ext.shape, y_ext.shape))

    # Set color map range and scale
    if z_range is None:
        vmin = z.min()
        vmax = z.max()
    else:
        vmin = min(z_range)
        vmax = max(z_range)

    if cmapscale == "log":
        z_min = z.min()
        if z_min <= 0:
            z_max = z.max()
            z = (z-z_min) + (z_max-z_min)*1e-4
            vmin = (vmin-z_min) + (z_max-z_min)*1e-4
            vmax = (vmax-z_min) + (z_max-z_min)*1e-4

    norm = get_cmapscale(cmapscale, vmin=vmin, vmax=vmax, clip=False)

    # return the matplotlib.collections.QuadMesh
    return ax.pcolormesh(x_ext, y_ext, z, cmap=cmap, norm=norm)
